/*** Giddes
 *     - Graphical Interface for Designing Discrete Event Systems
 *  S.Laurie Ricker Ph.D
 *
 *  Developed by I. Fraser & W. Fowler
 *
 *
 * reproduction without the express permission of Toba will
 * result in being taken outside and shot many times with a
 * very tiny gun.
 ***/

//*********************************************************
// Documentation completed January 19, 2002               *
//*********************************************************

/*
	This class is the BucketList that is used as the main collection
	for an FA's transitions. It does not store the transitions itself
	but stores the buckets that store the transitions.

	The List can access the entire collection of transitions from
	any one bucket. Any bucket can be added regardless if it is empty.
	When a new transition is to be added it is placed in the bucket
	that has the same name of the originating state of the transition.
	If no such bucket exists, it is created.

	The List also supports a comparision to another BucketList equals()
	To find if a particular edge exists it checks the originating state
	(bucket) and sees if the transition is there with the bucket's own
	searching routines.

*/
package umdes;

import java.util.Vector;
import java.lang.Exception;
import java.util.HashMap;
import java.util.Enumeration;

/**
 * A collection of Buckets that can only hold Transitions. This ADT is
 * to be used with the Fa object.
 *
 * @see UmdesAutomaton
 * @see Transition
 * @see Bucket
 */
public class BucketList{

	/**
	 * The collection of Buckets
	 */
	private HashMap<State, Bucket> buckets;

	/**
	 * Creates an empty BucketList
	 */
	public BucketList(){
	  buckets = new HashMap<State, Bucket>();
	}

	/**
	 * Copy Constructor, creates a new BucketList that is a copy of BucketList b
	 *
	 * @param bucketlist the BucketList to be copied
	 */
	public BucketList(BucketList bucketlist){
		buckets = new HashMap<State, Bucket>();
		Enumeration<Bucket> bucketCollection = bucketlist.getBuckets().elements();
		while(bucketCollection.hasMoreElements()){
			Bucket bucket = new Bucket(bucketCollection.nextElement());
			buckets.put(bucket.getState(),bucket);
		}
	}

  /**
   * Removes the entire contents of the BucketList.
   */
 	public void clearList(){
	  buckets.clear();
 	}

  /**
   * Sets the buckets of the list to those specified in the collection v
   * If the collection does not contain all Buckets, no changes are made to the present list
   *
   * @param v Contains the collection of buckets to set
   */
 	public void setBuckets(Vector<Bucket> v){
		HashMap<State, Bucket> bucketMap = new HashMap<State, Bucket>();

		Enumeration<Bucket> bucketCollection = v.elements();
		while(bucketCollection.hasMoreElements()){		
			Bucket curBucket = bucketCollection.nextElement();
			bucketMap.put(curBucket.getState(),curBucket);		
		}
		buckets = bucketMap;
	}

  /**
   * Sets the list to a copy of b
   *
   * @param b BucketList to copy
   */
	public void setList(BucketList b){
		setBuckets(b.getBuckets());
	}

  /**
   * Add all the buckets in v to the list (note: non-destructive addition)
   * No error checking is provided.
   *
   * @param v Collection containing the buckets to be added
   */
 	public void addAllBuckets(Vector<Bucket> v){
		 Enumeration<Bucket> e = v.elements();
		 Bucket temp;

		 while(e.hasMoreElements()){
		   temp = e.nextElement();
		   buckets.put(temp.getState(),temp);
		 }
	}

  /**
   * Adds the transition t to the appropriate bucket in the list
   *
   * @param t the Transition to add
   * @return True if the t is successfully added
   */
 	public boolean addTransition(Transition t){
		if(getBucket(t.getFromState())==null)
		  addBucket(t.getFromState());

		Bucket b = buckets.get(t.getFromState());
		if(b==null) return false;
		return b.addTransition(t);
	}

  /**
   * Adds a bucket to the list. If the bucket is already in the collection
   * another one will not be added
   *
   * @param s the added Bucket will represent State s
   */
 	public void addBucket(State state){
		 if(getBucket(state)!=null) return;
		 buckets.put(state, new Bucket(state)); 
	}

  /**
   * Removes a Bucket from the list specified by a certain State
   *
   * @param s The Bucket to be removed will have the same name as this State
   */
	public void removeBucket(State s) throws Exception{
		if(buckets.containsKey(s)){
			buckets.remove(s);
			Enumeration<Transition> trans = getAllTransitionsTo(s).elements();
			while(trans.hasMoreElements())removeTransition(trans.nextElement());
		}
		else throw new Exception("State is not contained in the Bucket List");
	}

  /**
   * Removes the specified Transition from the list
   *
   * @param t the Transition to be removed
   *
   * @return True if the specified Transition is found in the list and removed
   */
 	public boolean removeTransition(Transition t){
 		if(buckets.containsKey(t.getFromState())) // if bucket exist
			return getBucket(t.getFromState()).removeTransition(t); 
		else return false;
	}

  /**
   * Returns all transitions to a specified state. If no such Transitions
   * exist an empty collection is returned.
   *
   * @param s the specified state
   *
   * @return the collection of transitions to lead to State s. If there are
   * none than the colection will be empty
   */
 	public Vector<Transition> getAllTransitionsTo(State s){		
		Vector<Transition> result = new Vector<Transition>();
		Enumeration<Bucket> buckets = getBuckets().elements();
		while(buckets.hasMoreElements()){
			// get transitions for each bucket
			Enumeration<Transition> trans = buckets.nextElement().getTransitions().elements();
			while(trans.hasMoreElements()){
				Transition t = trans.nextElement();
				if(t.goesToState(s))result.add(t);
			}
		}
		return result;
	}

  /**
   * Returns true if the List is empty. There can be empty Buckets in a non-empty
   * BucketList. This method only checks if there are Buckets in the List
   *
   * @return True if the List is empty, ie: no Buckets(states). The Buckets
   * themselves maybe empty.
   */
 	public boolean isListEmpty(){
		return(buckets.isEmpty());
	}

  /**
   * Returns true if all the Buckets in the List are empty of Transitions
   *
   * @return True if the list contains no Transitions, that is the Buckets are empty
   */
  public boolean areBucketsEmpty(){
	  if(!buckets.isEmpty()){
		  Vector<Bucket> bucketVec = getBuckets();
		  for(int i = 0; i < buckets.size(); i++){
			  if(!bucketVec.elementAt(i).isEmpty()) return false;
		  }
	  }
	  return true ;
  }

  /**
   * Returns the number of Buckets in the List
   *
   * @return  number of buckets in the List
   */
 	public int getSize(){
 	 	return(buckets.size());
	}

  /**
   * Finds the Bucket representing the State s. If no such Bucket
   * is in the List, null is returned
   *
   * @param s the state to find
   *
   * @return the Bucket if found, otherwise null
   */
 	public Bucket getBucket(State s){
 		if(buckets.containsKey(s)) return(buckets.get(s));
 		else return null;
  }

  /**
   * Returns a collection of the Transitions that start at the State s
   * A Null return indicates no such Transitions exist
   *
   * @param s the originating state for the transitions
   *
   * @return the transitions from State s or null if there are no such Transitions
   */
 	public Vector<Transition> getTransitions(State s) throws Exception{
 		Bucket b=getBucket(s);
		if(b==null)return(null);
		return(b.getTransitions());
	}

  /**
   * Returns all the transitions that have an originating State s and a label RegExpr r
   * If no such Transitions exist a null is returned
   *
   * @param s the originating State
   * @param r the label of the Transitions
   *
   * @return the collection of Transitions or, if no such Transitions exist, null
   */
 	public Vector<Transition> getTransition(State s, RegExpr r){
		Bucket b=getBucket(s);
		if(b==null) return(null);
		return b.getTransition(r); 
	}

  /**
   * Returns a collection of the Buckets in the list
   * If there are none, the collection will be empty
   *
   * @return Vector - collection of the buckets
   */
 	public Vector<Bucket> getBuckets(){
 		Vector<Bucket> temp=new Vector<Bucket>(buckets.values());
 		return(temp);
	}

  /**
   * Tests if there is a Transition in the list with
   * the label of RegExpr r, regardless of where the transition starts or ends
   *
   * @param r a regular expression label
   *
   * @return True if there is a transition with a label r
   */
 	public boolean hasTransition(RegExpr r){
		Enumeration<Bucket> e = getBuckets().elements();		
		while(e.hasMoreElements()){ // for every Bucket
		  if(e.nextElement().hasTransitionWith(r))return true;
		}
		return false;
	}

  /**
   * More specific version of the method with the same name
   * The given State specifies an orginating State.
   *
   * @param state the orginating State
   * @param r a regular expression label
   *
   * @return True if there exists a Transition with st as an originating State
   * and r as a label
   */
	public boolean hasTransition(State state, RegExpr r){		
		Bucket b = getBucket(state);
		if(b==null) return false;
		return(b.hasTransitionWith(r)); 
	}

	/**
	 * Checks to see if State st has an Epsilon Transition leaving it.
	 *
	 * @param st the State to check
	 *
	 * @return True if the State st has an Epsilon Transition leaving it
	 */
	public boolean hasETransition(State state){
		Bucket b=getBucket(state); 
		if(b==null) return(false);
		return(b.hasETransition()); 
	}

  /**
   * Tests equality of Object obj with the BucketList
   *
   * @param obj the Object to test
   *
   * @return true if obj equals the BucketList
   */
  public boolean equals(Object obj){

	  Bucket bucket1, bucket2;
	  if(this==obj) return true;
	  if(obj==null) return false;

	  BucketList bucketlist;
	  if(obj instanceof BucketList){
		  bucketlist = (BucketList) obj;
	  } else return false;

	  if(getSize()!=bucketlist.getSize()) return false;

	  Vector<Bucket> bucketVec = bucketlist.getBuckets();

	  for(int i = 0; i < bucketVec.size(); i++){
		  bucket1 = bucketVec.get(i);
		  if(!buckets.containsKey(bucket1.getState())) return false;
		  bucket2 = buckets.get(bucket1.getState());
		  if(!bucket1.equals(bucket2)) return false;
	  }

	  return true;
  }

  /**
   * The returned collection specifies all possibles States that can be reached from State st
   * with an event s. If no such Transitions are possible then a null is returned
   *
   * @param state the originating State
   * @param event the event that occured
   *
   * @return the collection of states that can be reached by this event,
   * null if there is no such transitions
   *
   * @throws Exception on String input that is not a regular expression
   */
   public Vector<State> doTransition(State state, String event) throws Exception{

	   if(!RegExpr.isRegExpression(event)){
		   throw new Exception("Invalid Input");
	   }
		
	   Bucket b = getBucket(state);
	   if(b==null)  return(null);

	   return b.doTransition(state , new RegExpr(event));
   }

  /**
   * The returned collection specifies all possibles States that can be reached from State st
   * with an event r. If no such Transitions are possible then a null is returned
   *
   * @param st the originating State
   * @param r the regular expression event that occured
   *
   * @return the collection of states that can be reached by this event,
   * null if there is no such transitions
   */
  public Vector<State> doTransition(State state, RegExpr r){

	  Bucket b = getBucket(state);
	  if(b==null) return(null);

	  return b.doTransition(state,r);
  }

  /**
   * Edits an existing Transition t, giving it a new from State, a new to State and a new label
   * No changes are made if Transition t does not exist. An Exception is thrown in this case.
   *
   * @param t the Transition to be edited
   * @param newFrom the State to be made the new From State
   * @param newTo the State to be made the new To State
   * @param newLabel the new Label
   *
   * @throws Exception if t does not exist in the BucketList
   */
  public void editTrans(Transition t, State newFrom, State newTo, RegExpr newLabel) throws Exception{
	  Bucket b = getBucket(t.getFromState());

	  if(b==null) throw new Exception("No such Transiton to be edited");

	  b.removeTransition(t);
	  addTransition(new Transition(newFrom,newTo,new TransitionLabel(newLabel)));
  }

  /**
   * Rename the State s with the name newname. If the State does not exist
   * an Exception is thrown.
   *
   * @param s the State to be renamed
   * @param newname the new name for the State
   *
   * @throws Exception if State s does not exist in the BucketList
   */
  public void renameState(State state, String newname) throws Exception{
	  
	  State newState = new State(newname);
	  Vector<Transition> trans = getAllTransitionsTo(state);
	  Enumeration<Transition> e;
    
	  // rename all the transition into the state
	  if(!trans.isEmpty()){
		  e = trans.elements();
		  while(e.hasMoreElements()){
			  Transition t = e.nextElement();
			  editTrans(t, t.getFromState(), newState, t.getLabel());
		  }
	  }
	  Bucket b = getBucket(state);
	  if(b==null) throw new Exception("State does not exist");
    
	  b.renameBucket(newState);
	  buckets.remove(state);
	  buckets.put(newState,b);  
  }

  /**
   * Test for non-determinism in the BucketList. All Buckets are checked
   *
   * @return True if there are non-deterministic Transitions in the BucketList
   */
  public boolean checkNondeter(){
    Enumeration<Bucket> e = getBuckets().elements();
    while(e.hasMoreElements()){
    	Bucket b = e.nextElement();
    	if(b.isNondeter()) return(true);
    }
    return(false);
  }

  /**
   * Test whether there exists a Transition from State a to State b
   *
   * @param a the Originating State
   * @param b the Visted State
   *
   * @return True if there is at least one Transition from State a to State b
   */
  public boolean isThereAToB(State a, State b){
    Bucket bucket = getBucket(a);
    if(bucket==null) return false;

    return(bucket.isToB(b)); 
  }
  
};
