package umdes;

import giddes.app.GuiConstants;
import giddes.app.GuiToolBar;
import giddes.app.StatusNotifier;
import giddes.app.workspace.GuiStatusBar;
import giddes.commands.commands.BatchOpen;
import giddes.commands.commands.BatchSaveAs;
import giddes.commands.commands.Help;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.JToolBar;
import javax.swing.ScrollPaneConstants;

/**
 * Very similar to the GuiBatchWindow used in desuma, except it is more flexible
 * to allow the user to enter whatever is needed
 * @author John O'Brien
 *
 */
@SuppressWarnings("serial")
public class InputWindow extends JDialog implements StatusNotifier{
	/** Constant that represents the width of the window. */
	private static final int BWidth = 480;
	/** Constant that represents the height of the window. */
	private static final int BHeight = 350;
	/** Constant that represents the title of the window. */
	private static final String BTitle = GuiConstants.APPLICATION_TITLE + " Input Window";

	/** Constant that represents the color of selected text in the command area. */
	private static final Color DEF_SELECTION_TEXT_COLOR = Color.black;
	/** Constant that represents the color of the selection highlight in the command area. */
	private static final Color DEF_SELECTION_COLOR = new Color(206, 206, 206);
	/** Constant that represents the color of the header in the command area. */
	private static final Color HEADER_BG_COLOR = new Color(206, 206, 206);
	/** Constant that represents the number of columns in the command area. */
	private static final int DEF_COMMAND_COLS = 12;
	/** Constant that represents the number of rows in the command area. */
	private static final int DEF_COMMAND_ROWS = 20;
	
	/** Text area where the commands will be typed. */
	private JTextArea commands;
	/** Text to be put in the status bar. */
	private JLabel status_text;
	/** Button that will submit current commands to the system. */
	private JButton submit;
	/** Button that will cancel the window without submitting. */
	private JButton cancel;
	/** Header text to be placed above the command area. */
	private JLabel head_text;
	/** Buttons that will open files to the batch window, save commands from the batch window, will clear the text area, and will bring up the batch help window. */
	private JButton open, save, clear, help;
	
	/** The status bar from the frame. */
	private GuiStatusBar status_bar;
	
	private boolean submitPressed = false;

	/**
	 * Default constructor that creates a Batch window.
	 * The dialog is created as modal, with <code>BTitle</code> as the title of the window.
	 * @param parentFrame The frame that this batch window should appear on top of.
	 * @throws FileNotFoundException If there is an error reading in the images for the tool bar.
	 */
	public InputWindow(){
		//super(parentFrame, BTitle, false); // true defaults this dialog to modal.
		setTitle(BTitle);
		setResizable(true);
		setPreferredSize(new Dimension(BWidth, BHeight));
		setDefaultCloseOperation(javax.swing.JFrame.HIDE_ON_CLOSE);
		
		//registerStatus(parentFrame.getStatusBar());
		addWindowListener(new WindowAdapter() {
			public void windowClosing(final WindowEvent event) {				
				System.exit(0);
			}
		});

		final JToolBar toolbar = initToolbar();
		final JPanel main_window = initMainWindow();
		final JPanel status_bar = initStatusBar();
		
		getContentPane().setLayout(new BorderLayout());
		getContentPane().add(toolbar, BorderLayout.NORTH);
		getContentPane().add(main_window, BorderLayout.CENTER);
		getContentPane().add(status_bar, BorderLayout.SOUTH);
			
		pack();

		setLocationRelativeTo(null);
	}

	/**
	 * Sets the text of the header.
	 * @param text The text that should be set to the header.
	 */
	public void setHeaderText(final String text){
		head_text.setText(text);
	}
	public String getText(){
		return commands.getText();
	}
	
	/**
	 * Sets the text in the status bar.
	 * @param text The text that should be set to the status.
	 */
	public void setStatusText(final String text) {
		status_text.setForeground(Color.GRAY); // Set text to a grayed color.
		status_text.setText(text);
	}

	/**
	 * Sets the text in the status bar to the specified error message.  The text
	 * will be displayed in red.
	 * @param error the error message to display
	 */
	public void setStatusError(final String error) {
		status_text.setForeground(Color.RED);
		status_text.setText(error);
	}

	/**
	 * Initializes the toolbar.
	 * @return A JToolBar that can be used as the toolbar.
	 * @throws FileNotFoundException If the toolbar button icons cannot be found.
	 */
	protected JToolBar initToolbar(){
		final JToolBar toolbar = new JToolBar();
		toolbar.setFloatable(false);

		final ActionListener clearCommands = new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				commands.setText("");
			}
		};

		open  = GuiToolBar.createButton(GuiToolBar.ICON_OPEN, GuiConstants.APP_OPEN, true, new BatchOpen());
		save  = GuiToolBar.createButton(GuiToolBar.ICON_SAVE, GuiConstants.APP_SAVE, true, new BatchSaveAs());
		clear = GuiToolBar.createButton(GuiToolBar.ICON_ERASER, "Clear Commands Area", true, clearCommands);
		help  = GuiToolBar.createButton(GuiToolBar.ICON_HELP, GuiConstants.APP_HELP, true, new Help("help"));

		toolbar.add(open);
		toolbar.add(save);
		toolbar.add(clear);
		toolbar.add(Box.createHorizontalGlue());
		toolbar.add(help);
		
		return toolbar;
	}
	
	/**
	 * Initializes the main window
	 * @return A JPanel that can be used as the main window.
	 */
	protected JPanel initMainWindow(){
		final JPanel main_window = new JPanel();
		main_window.setLayout(new BoxLayout(main_window, BoxLayout.Y_AXIS));
		main_window.setBorder(BorderFactory.createEtchedBorder(1));

		// Header text
		final JPanel headP = new JPanel();
		headP.setLayout(new BorderLayout());
		headP.setBackground(HEADER_BG_COLOR);
		head_text = new JLabel("Enter " + GuiConstants.APPLICATION_TITLE + " commands below");
		head_text.setHorizontalAlignment(JLabel.CENTER);
		head_text.setHorizontalTextPosition(JLabel.CENTER);
		head_text.setVerticalAlignment(JLabel.CENTER);
		head_text.setVerticalTextPosition(JLabel.CENTER);
		headP.setMaximumSize(new Dimension(BWidth, 100));
		headP.setPreferredSize(new Dimension(BWidth, 20));
		headP.add(head_text, BorderLayout.CENTER);
		headP.add(head_text);
		
		// Command Area
		commands = new JTextArea(DEF_COMMAND_COLS, DEF_COMMAND_ROWS);
		commands.setLineWrap(false);
		commands.setEditable(true);
		commands.setSelectionColor(DEF_SELECTION_COLOR);
		commands.setSelectedTextColor(DEF_SELECTION_TEXT_COLOR);
		final JScrollPane commandsP = new JScrollPane(commands);
	    commandsP.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS);
	    commandsP.setViewportBorder(BorderFactory.createLineBorder(Color.gray));
		
		// Submit and cancel buttons
		final JPanel button_panel = new JPanel();
		button_panel.setLayout(new BoxLayout(button_panel, BoxLayout.X_AXIS));
		button_panel.add(Box.createHorizontalGlue());
		submit = new JButton("Submit");
		final ActionListener submitCommands = new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				InputWindow.this.setVisible(false);
				submitPressed = true;
			}
		};
		submit.addActionListener(submitCommands);

		cancel = new JButton("Cancel");
		final ActionListener hideWindow = new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				// Hide the window
				InputWindow.this.setVisible(false);
				System.exit(0);
			}
		};
		cancel.addActionListener(hideWindow);

		button_panel.add(submit);
		button_panel.add(cancel);
		
		main_window.add(headP);
		main_window.add(commandsP);
		main_window.add(button_panel);
		
		return main_window;
	}
	
	/**
	 * Initializes the status bar.
	 * @return A JPanel that can be used as the status bar.
	 */
	protected JPanel initStatusBar(){
		final JPanel sb = new JPanel();
		sb.setLayout(new BoxLayout(sb, BoxLayout.X_AXIS));
		status_text = new JLabel("Ready");
		status_text.setForeground(Color.GRAY); // Set text to a grayed color.
		sb.add(status_text);
		sb.add(Box.createHorizontalGlue());
		return sb;
	}
		
	/**
	 * Reads a text file into the batch window.
	 * @param txtFile File to be read in.
	 */
	public void readFromTextFile(final File txtFile)
	{
		if (txtFile != null)
		{
			BufferedReader input = null;
			try{
				commands.setText("");
				input = new BufferedReader(new FileReader(txtFile));
				String command = input.readLine();
				
				while (command != null)
				{
					commands.append(command.replaceAll("\\s+$","") + "\n");
					command = input.readLine();
					
				}
			}
			catch (final IOException e)
			{
				System.out.println(e.toString());
				e.printStackTrace();
			}
			finally
			{
				if (input != null)
				{
					try{
						input.close();
					}
					catch (final IOException e)
					{
						System.out.println(e.toString());
						e.printStackTrace();
					}
				}
			}
		}
	}
	
	/**
	 * Saves the commands entered in the batch window to a specified text file.
	 * @param txtFile File to save the batch window commands.
	 */
	public void saveToTextFile(final File txtFile)
	{
		if (txtFile != null){
			BufferedWriter output = null;
			try{
				output = new BufferedWriter(new FileWriter(txtFile));
				output.write(commands.getText(), 0, commands.getText().length());
				output.flush();
			}
			catch (final IOException e)
			{
				System.out.println(e.toString());
				e.printStackTrace();
			}
			finally
			{
				if (output != null)
				{
					try{
						output.close();
					}
					catch (final IOException e)
					{
						System.out.println(e.toString());
						e.printStackTrace();
					}
				}
			}
		}
	}

	/**
	 * Method that registers the current status bar.
	 * @param statusBar The status bar to be registered.
	 */
	public void registerStatus(final GuiStatusBar statusBar){
		status_bar = statusBar;
	}
	
	/**
	 * Method that sets the status text of the status bar.
	 * @param text The text to set in the status bar.
	 */
	public void notifyText(final String text){
		status_bar.setStatus(text);
	}

	/**
	 * Displays an error to the user, using the standard JOptionPane error message dialog.
	 *
	 * @param errMessage the error message to display
	 */
	public void showError(final String errMessage) {
		JOptionPane.showMessageDialog(this, errMessage, GuiConstants.APPLICATION_TITLE + " Batch Commands Error", JOptionPane.ERROR_MESSAGE);
	}

	public void setSubmitPressed(final boolean submitPressed) {
		this.submitPressed = submitPressed;
	}

	public boolean isSubmitPressed() {
		return submitPressed;
	}
}
	


