/*** Giddes
 *     - Graphical Interface for Designing Discrete Event Systems
 *  S.Laurie Ricker Ph.D
 *
 *  Developped by I. Fraser & W. Fowler
 *
 *
 * reproduction without the express permission of Toba will
 * result in being taken outside and shot many times with a
 * very tiny gun.
 ***/

//*********************************************************
// Documentation completed Jan, 27 2002                   *
//*********************************************************

package umdes;

import giddes.xml.automaton.Event;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
import java.util.Vector;

import javax.swing.JFrame;

//import javax.swing.JOptionPane;


/**
 * A Finite Automata is described by a State set, Transition Set,
 * a set of Terminating States and an initial State.
 *
 * @see State
 * @see Transition
 * @see BucketList
 */
public class UmdesAutomaton
{
  //***************************************************************
  // constants
  //***************************************************************
  /**
   * An Exception Message: "No Initial State Selected"
   */
  public static final String NOINITIAL="No Initial State Selected";

  //****************************************************************
  // protected members
  //****************************************************************

  /**
   * State set
   */
  private Vector<State> states;
  /**
   * Transition set
   */
  private BucketList events;
  /**
   * set of Marked(Terminating) States
   */
  private Vector<State> markedStates;
  /**
   * The alphabet of symbols that are recognized by the Automata
   */
  private Vector<RegExpr>  alphabet;
  /**
   * The set of uncontrollable events
   */
  private Vector<RegExpr> uncontrol;
  /**
   * The set of unobservable events
   */
  private Vector<RegExpr> unobserv;

  /**
   * The initial State of the Automata
   */
  private State initial;

  /**
   * A list of inadmissable events (ie all those which need to be disabled)
   */
  private Vector<Transition> inadmissable;
  
  /**
   * The specified states that can never be possible at the same time
   */
  private T_spec t_spec;   //JO'B  5/20/09
  private T_Conf t_conf;
  

  private Vector<RegExpr> trans_order; 
  private Set<Qc> possibleStateSets;
  private Vector<Window> windowPartitions;
  
  private BufferedWriter outOnlineMinSen;
  private Vector<String> EventOrder; 
  private Vector<String> totalOrder;
  private Qc stateSet;
  
  /**
   * Creates an empty machine
   */
  public UmdesAutomaton(){
  	states = new Vector<State>();
  	markedStates = new Vector<State>();
  	events = new BucketList();
  	alphabet = new Vector<RegExpr>(); //<-at the moment the alphabet only supports single RegExprs
  	uncontrol=new Vector<RegExpr>();
  	unobserv=new Vector<RegExpr>();
  	inadmissable = new Vector<Transition>();
  	t_spec = new T_spec();
  	setStateSet(new Qc());
  	setTrans_order(new Vector<RegExpr>());
  	possibleStateSets = new HashSet<Qc>();
  	windowPartitions = new Vector<Window>();
  }
  public UmdesAutomaton(giddes.xml.automaton.Automaton automaton){
	  	states = new Vector<State>();
	  	markedStates = new Vector<State>();
	  	events = new BucketList();
	  	alphabet = new Vector<RegExpr>(); //<-at the moment the alphabet only supports single RegExprs
	  	uncontrol=new Vector<RegExpr>();
	  	unobserv=new Vector<RegExpr>();
	  	inadmissable = new Vector<Transition>();
	  	t_spec = new T_spec();
	  	setStateSet(new Qc());
	  	setTrans_order(new Vector<RegExpr>());
	  	possibleStateSets = new HashSet<Qc>();
	  	windowPartitions = new Vector<Window>();
	  	
		// Add states
		for (giddes.xml.automaton.State s: automaton.getState()) {
			try {
				State state = new State(s.getName());
				states.add(state);
				if (s.isInitial()==Boolean.TRUE) setInitial(state);
				if (s.isMarked() ==Boolean.TRUE)  markState(state);
			} catch(Exception e) {
				e.printStackTrace();
				return;
			}
		}
		// Add transitions
		for (giddes.xml.automaton.Transition t: automaton.getTransition()) {
			
			State from = getState(t.getFrom());
			State to   = getState(t.getTo());
			if (from == null || to == null) {
				return;
			}
			try {
				Transition trans = new Transition(from, to, new RegExpr(t.getName()));
				addTransition(trans);
			} catch(Exception e) {}
		}

		for (Event e : automaton.getEvent()) {
			if(e.isControllable()!=Boolean.TRUE) addUncontrolEvent(new RegExpr(e.getName()));
			if(e.isObservable()  !=Boolean.TRUE) addUnobserveEvent(new RegExpr(e.getName()));
		}
  
  }

  /**
   * Sets the initial State of the machine. Throws Exception
   * if the State does not exist in the machine.
   *
   * @param s the initial state
   *
   * @throws Exception if the State is not a part of the machine
   */
  public void setInitial(State s) throws Exception{
  	if(hasState(s))
  	  initial = s;
  	else
  	  throw new Exception("State does not Exist");
  }

  /**
   * Sets the initial State of the machine, specified by the String s,
   * the name of the State. Throws Exception
   * if the State does not exist in the machine.
   *
   * @param s the initial state
   *
   * @throws Exception if the State is not a part of the machine
   */
  public void setInitial(String s) throws Exception{
    State temp = new State(s);
	setInitial(temp);
  }

  public Vector<Transition> getInadmissable(){
  	if(inadmissable == null)
  		return null;

  	return(new Vector<Transition>(inadmissable));
  }

  public void setInadmissable(Vector<Transition> list){
  	inadmissable=list;
  }
 
  /**
   * Returns the State set of the Machine
   *
   * @return The States of the Machine
   */
  public Vector<State> getStates(){
    return(new Vector<State>(states));
  }

  /**
   * Returns the BucketList containing all the Transitions of the machine
   *
   * @return all the Transitions of the machine
   */
  public BucketList getEvents(){
    return(new BucketList(events));
  }
  
  
  /**
   * returns all of the Observable Events
   * @return : A vector of observable RegExpr's in String form
   */
  public Vector<String> getObservableEvents(){  
	  if(EventOrder==null){
		  EventOrder = new Vector<String>();
		  for(int i = 0; i< alphabet.size(); i++){
			  RegExpr r = new RegExpr(((TransitionLabel) alphabet.get(i)).getEvent());
			  if(!isUnobserv(r))  EventOrder.add(r.expression);
		  }
	  }
	  return EventOrder;	  
  }
  public void resetEventOrder(){
	  EventOrder = null;
	  totalOrder = null;
  }
  
  public Vector<RegExpr>  getAlphabet(){
	  Vector<RegExpr> v = new Vector<RegExpr> ();
	  for(int i = 0; i< alphabet.size(); i++){
		  v.add(new RegExpr(((TransitionLabel) alphabet.get(i)).getEvent()));
	  }
	  return v;
  }


  /**
   * Returns the initial State of the machine
   *
   * @return the initial State
   */
  public State getInitial(){
    if(initial == null)
      return null;
    return(new State(initial));
  }

  public boolean isIllegal(Transition t){
	return inadmissable.contains(t);

  }


  /**
   * Removes the RegExpr r from the alphabet of the machine
   * True return indicates that r was successfully removed.
   *
   * @param r the RegExpr to remove from the alphabet
   *
   * @return True if r is successfully removed
   */
  public boolean removeFromAlphabet(RegExpr r){
  	if(alphabet.contains(r))
  		return(alphabet.remove(r));
  	else
  	  return(false);
  }

  /**
   * Adds a RegExpr to the alphabet of the machine, if it is
   * not already present in the alphabet
   *
   * @param r the RegExpr to add to the alphabet
   */
  public void addToAlphabet(RegExpr r){
    if(!alphabet.contains(r))
      alphabet.add(r);

  	}


  /**
   * Specifies the RegExpr r as an Uncontrollable event for this machine (Controller)
   *
   * @param r the event that is Uncontrollable by this machine
   */
  public void addUncontrolEvent(RegExpr r){
    if(!uncontrol.contains(r))
      uncontrol.add(r);
  }

  /**
   * Specifies the RegExpr r as a  controllable event for this machine (Controller)
   * if it was previously specified as uncontrollable by a addUncontrollable event
   * method call.
   *
   * @param r the event that is Uncontrollable by this machine
   */
  public void removeUncontrolEvent(RegExpr r){
  	boolean found=false;
  	int i;
	for(i=0; !found&&i<uncontrol.size(); i++){
		RegExpr temp = (RegExpr)uncontrol.elementAt(i);
		if(temp.equals(r)){
			found=true;
  }
	}
	i--;
	if(!found){
		return;
	}
	else{
		uncontrol.remove(i);
	}
  }

  /**
   * Specifies a RegExpr as an unobservable event for this machine(Observer)
   *
   * @param r the RegExpr that is unobservable by this machine
   */
  public void addUnobserveEvent(RegExpr r){
    if(!unobserv.contains(r))
      unobserv.add(r);
  }

  /**
   * Specifies a RegExpr as an Observable event for this machine(Observer)
   * This negates a addUnobservEvent(RegExpr) method call.
   *
   * @param r the RegExpr that is observable by this machine
   */
  public void removeUnobservEvent(RegExpr r){
  	boolean found=false;
  	int i;
	for(i=0; !found&&i<unobserv.size(); i++){
		RegExpr temp = (RegExpr)unobserv.elementAt(i);
		if(temp.equals(r)){
			found=true;
		}
	}
	i--;
	if(!found){
		return;
	}
	else{
		unobserv.remove(i);
	}
  }

  /**
   * Test whether the RegExpr r is an uncontrollable event for this machine
   *
   * @param r the RegExpr to be tested for controllability
   *
   * @return True if RegExpr r is uncontrollable by this machine
   */
  public boolean isUncontrol(RegExpr r){
	return uncontrol.contains(r);
  }

  /**
   * Test whether the RegExpr r is an unobservable event for this machine
   *
   * @param r the RegExpr to be tested for observability
   *
   * @return True if RegExpr r is unobservable by this machine
   */
  public boolean isUnobserv(RegExpr r){
	return unobserv.contains(r);
  }



  /**
   * Adds a new State to the machine with name s
   * s must be a valid State name without any of these
   * characters: ',' '{' '}' '(' ')'
   *
   * @param s the name for the new State
   *
   * @throws Exception if s is not a valid State name
   */
  public void addState(String s) throws Exception{
  	State temp;

  	temp= new State(s);
	addState(temp);
  }

  /**
   * Adds a the State s to the machine
   * If the name of the State is not unique to the machine
   * an Exception is thrown
   *
   * @param s the new State to add
   *
   * @throws Exception if the new State's name is not unique to the machine
   */
  public void addState(State s) throws Exception{
    if(hasState(s))
      throw new Exception("Not a Unique State Name");
  	states.addElement(s);
      // add the bucket
    events.addBucket(s);

  }

  /**
   * Adds a the State s to the machine
   * If the name of the State is not unique to the machine
   * an Exception is thrown
   *
   * @param s the new State to add
   *
   * @throws Exception if the new State's name is not unique to the machine
   */
  public void addIllegalTransition(Transition t){
    inadmissable.add(t);
  }

  /**
   * Adds a the State s to the machine
   * If the name of the State is not unique to the machine
   * an Exception is thrown
   *
   * @param s the new State to add
   *
   * @throws Exception if the new State's name is not unique to the machine
   */
  public void removeIllegalTransition(Transition t){
    inadmissable.remove(t);
  }

  /**
   * Creates a new Transition with the originating State f,
   * the to State t, and the label r, then adds the Transition
   * to the machine. If it is already there, no changes are made
   * and the return is false.
   *
   * @return if the Transition is successfully added
   */
  public boolean addTransition(State f, State t, RegExpr r){
  	Transition temp;
  	temp = new Transition(f,t,r);

	return addTransition(temp);
    }

  /**
   * Adds the Transition t to the machine. If it is already there,
   * no changes are made and the return is false.
   *
   * @param t the Transition to be added
   *
   * @return if the Transition is successfully added
   */
  public boolean addTransition(Transition t){

  	if(events.addTransition(t)){
  	  // add the label to the alphabet
  	  addToAlphabet(t.getLabel());
  	  return(true);
    }
    else
      return(false);
  }


  /**
   * Marks the State with the name s
   * An Exception is thrown if no such State exists in the machine
   *
   * @param s the name to the State to mark
   *
   * @throws Exception if s does not specify a State of this machine
   */
  public void markState(String s) throws Exception{
  	State temp;

  	temp=getState(s);

	markState(temp);
  }

  /**
   * Marks the provided State s
   * An Exception is thrown if no such State exists in the machine
   *
   * @param s the State to mark
   *
   * @throws Exception if s is not a State of this machine
   */
  public void markState(State s) throws Exception{
  	if(s==null||!hasState(s))
  	  throw new Exception("State Does Not Exist");

  	if(isMarkedState(s))
  	  return;

  	markedStates.addElement(s);
  }

  /**
   * Unmarks the provided State s from the machine, undoing a Fa.markState method call
   * If the State s is not marked no changes are made
   *
   * @param s the State to unmark
   */
  public void unmarkState(State s){
  	if(isMarkedState(s)){
  		markedStates.remove(s);
  	}
  }

  /**
   * Removes the State with the name s from the machine. Removing a State
   * removes all the Transitions associated with it. An Exception is thrown
   * if the machine Transition data is corrupted  or if s is the initial State.
   * A further description will be given by the Exceptions message.
   *
   * @param s the name of the State to remove
   *
   * @throws Exception if the machine's data is corrupted or if s is the initial State
   */
  public void removeState(String s) throws Exception{
  	State temp;

    temp=getState(s);
    removeState(temp);
  }

  /**
   * Removes the State specified from the machine. Removing a State
   * removes all the Transitions associated with it. An Exception is thrown
   * if the machine Transition data is corrupted or if s is the initial State.
   * A further description will be given by the Exceptions message
   *
   * @param s the State to remove
   *
   * @throws Exception if the machine's data is corrupted or if s is the initial State
   */
  public void removeState(State s) throws Exception{
    if(hasState(s)){ // check that the State is present
	  	if(s!=null){
	      states.removeElement(s); // remove State

	      if(isMarkedState(s))
	      	markedStates.removeElement(s); // remove it from the marked set if necessary

      if(initial!=null){  // if there is an initial state specified
        if(initial.equals(s))
          initial=null;
      }

      try{
	      	events.removeBucket(s);	 // clean up Transitions into and out of State
      }
      catch(Exception e){
        throw new Exception(e.getMessage());
      }

    }
  }
  }


  /**
   * Removes the Specified Transition from the machine and also removes
   * the label from the machine's alphabet if it is the only Transition
   * with that label in the machine
   */
  public void removeTransition(Transition t){

  	events.removeTransition(t);
  	if(!events.hasTransition(t.getLabel())){
  	  this.removeUncontrolEvent(t.getLabel());
  	  this.removeUnobservEvent(t.getLabel());
  	  removeFromAlphabet(t.getLabel());
    }
  }

  /**
   * Returns the State in the machine that has the name s
   * If no State has the name s a null is returned
   *
   * @param s the name of the State
   *
   * @return the State with the name s or if it can't be found a null
   */
  public State getState(String s){
	  Enumeration<State> i =states.elements();
	  while(i.hasMoreElements()){
		  State state =i.nextElement();
		  if(state.equals(new State(s))) return(state);
	  }
	  return(null);
  }

  /**
   * Tests to see if the String s is State name in this machine
   *
   * @param s the String to test
   *
   * @return True if s is a State name in this machine
   */
  public boolean hasState(String s){
  	State temp = new State(s);
  	return hasState(temp);
    }

  /**
   * Tests to see if the State s is a State in this machine
   *
   * @param s the State to test
   *
   * @return True if s is a State in this machine
   */
  public boolean hasState(State s){
  	Enumeration<State> state = states.elements();
  	while(state.hasMoreElements()){
  		if(state.nextElement().equals(s)) return(true);
    }
  	return(false);
  }

  /**
   * Tests to see if the State s is a marked State of this machine
   *
   * @param s the State to test
   *
   * @return True if s is a marked State of this machine
   */
  public boolean isMarkedState(State s){
  	Enumeration<State> marked = markedStates.elements();
  	while(marked.hasMoreElements()){
  		if(s.equals(marked.nextElement()))return(true);
    }
  	return(false);
  }

  /**
   * Tests to see if the String s the name of a marked State of this machine
   *
   * @param s the String to test
   *
   * @return True if s name's a marked State of this machine
   */
  public boolean isMarkedState(String s){
  	State temp = new State(s);
  	return isMarkedState(temp);
    }



/**
 * this function will enters the window partitions and finds which 
 * order to analyze them it is possible to have window sizes larger
 * than 1, but currently that is all that is supported
 * @param n window size
 * @param parent used to to properly display the window ordering 
 */
  public void initializeWindowPartition(int n) throws IOException{

	  JFrame window = new JFrame();
	  if(n==1){
		  //windowPartitions.clear();
		  RegExpr r;
		  if(windowPartitions.isEmpty()){
			  
			  for(int i = 0; i < states.size(); i++){
				  Vector<Transition> trans = new Vector<Transition>();
				  try{
					  trans = events.getTransitions((State) states.get(i));
				  } catch (Exception e) {}
				  if(trans==null || trans.size()==0) continue;
				  for(int j = 0; j < trans.size(); j++){
					  r = ((Transition)trans.get(j)).getLabel().getEvent();
					  if(!isUnobserv(r)){
						  windowPartitions.add(new Window(states.get(i), n, new Vector<String>(), r));
					  }
				  }
			  }
		  }
	      EventList ev = new EventList(windowPartitions, window, true);
	   	  if(ev.getEventOrder()==null||ev.getEventOrder().size()==0) throw new IOException();

	   	  windowPartitions = ev.getEventOrder();
	   	     	  
	   	  t_conf = new T_Conf(this, ev.getEventOrder(), windowPartitions);
 		  
	  } 	  
  }
  
  /**
   * Computes the minimum sensor policy to keep the automata diagnosable. 
   * It uses a cluster verifier to determine the diagnosablility as the windows
   * are being removed 
   * @param failEvents a vector of RegExpr that are failure events
   * @param workDir the current working directory (needed by cluster verifier)
   * @return a Vector of transitions that should be turned off
   */
  public Vector<Transition> computeMinSenDiag(Vector<RegExpr> failEvents, String workDir){
	  
	  String filename = "min_sen_diag.fsm";
	  Vector<Transition> all_trans = getAllTransitions(); 
	  boolean inD = false;

	  //T_Conf local_tconf = t_conf;
	  for(int i = 0; i < t_conf.winSize(); i++){

		  inD = false;
		  if(t_conf.isWindowOff(t_conf.getWindow(i))) continue;
		  
		  t_conf.turnOffWindow(t_conf.getWindow(i));
		  t_conf.expandTConf(this);
		  
		  
		  for(int j = 0; j < t_conf.sensorOffSize(); j++){
			  if(t_conf.isInD(t_conf.getSensorOff(j))){
				  t_conf.addToD(t_conf.getWindow(i));
				  t_conf.resetSensors();				  
				  inD = true;
			  }
		  }
		  if(inD)continue;
		  
		  String diagnose = runCVerify(t_conf.getPolicy(this), all_trans, workDir, filename, failEvents, true);
		  
		  if(diagnose.equals("true")) t_conf.keepSensorChanges();
		  else{
			  t_conf.addToD((Window) t_conf.getWindow(i));
			  t_conf.resetSensors();
		  }
	  
	  }
	  return t_conf.getPolicy(this).get(0);
	  
  }
  
  /**
   * Returns all the Transitions from the State s
   *
   * @param s the originating State of the desired Transitions
   *
   * @return the Transitions from the State s
   */
  public Vector<Transition> getStatesTransitions(State s) throws Exception{
  	return(events.getTransitions(s));
  }
  
  /**
   * will find live lock and dead lock and return the states in a Vector.  
   * If there is an error it will return a vector with the first element 
   * being a string describing the error.
   */
  public Vector<?> findBlocking(){
	  
	  if(markedStates==null||markedStates.size()==0){
		  Vector<String> v =  new Vector<String>();
		  v.add("There are no marked states");
		  return v;
	  }
	  Vector<State> seenStates = new Vector<State>();
	  Vector<State> goodStates = new Vector<State>();  
	  Vector<Transition> trans = new Vector<Transition>();       //will hold transitions
	  
	  Vector<State> suspectStates = new Vector<State>();
	  
	  State cur = getInitial();
	  
	  //put all states that can lead to the marked state in goodStates
	  for(int i = 0; i < markedStates.size(); i++){
		  trans = events.getAllTransitionsTo((State) markedStates.get(i));
		  if(!goodStates.contains((State) markedStates.get(i)))
			  goodStates.add((State) markedStates.get(i)); 
		  for(int j = 0; j < trans.size(); j++){
			  if(!goodStates.contains(((Transition) trans.get(j)).from))
				  goodStates.add(((Transition) trans.get(j)).from); 
		  }
	  }
	  for(int i = 0; i < goodStates.size(); i++){
		  trans = events.getAllTransitionsTo((State) goodStates.get(i));
		  for(int j = 0; j < trans.size(); j++){
			  if(!goodStates.contains(((Transition) trans.get(j)).from))
				  goodStates.add(((Transition) trans.get(j)).from); 
		  }		  
	  }
	  
	  if(!goodStates.contains(cur)){
		  Vector<String> v =  new Vector<String>();
		  v.add("No string from the initial states can reach any marked state");
		  return v;
	  }
	  
	  //put all states that can be reached from the initial states but cannot 
	  //make it to the marked state in suspectStates.
	  seenStates.add(cur);
	  for(int i = 0; i < seenStates.size(); i++){
		  try{
			  trans = events.getTransitions((State) seenStates.get(i)); 
		  } catch(Exception e) { continue;}
		  for(int j = 0; j < trans.size(); j++){
			  if(!seenStates.contains(((Transition) trans.get(j)).to))
				  seenStates.add(((Transition) trans.get(j)).to);
			  else continue;
			  if(!goodStates.contains(((Transition) trans.get(j)).to))
				  suspectStates.add(((Transition) trans.get(j)).to);
		  }		  
	  }
	  for(int i = 0; i < states.size(); i++){
		  if(!goodStates.contains((State) states.get(i))&&!seenStates.contains((State) states.get(i)))
			  suspectStates.add((State) states.get(i));
	  }
	  //suspect states is sufficient to find which will never get to the marked states
	  return suspectStates;
	  
  }

  
  public Vector<State> findReachableStates(State in){
	  
	  Vector<State> statesOut = new Vector<State>();
	  Vector<Transition> trans = new Vector<Transition>();
	  statesOut.add(in);
	  
	  for(int i = 0; i < statesOut.size(); i++){
		  try{
			  trans = events.getTransitions((State) statesOut.get(i)); 
		  } catch(Exception e) { continue;}
		  for(int j = 0; j < trans.size(); j++){
			  if(!statesOut.contains(((Transition) trans.get(j)).to))
				  statesOut.add(((Transition) trans.get(j)).to);
		  }
	  }
	  return statesOut;
	  
  }
  public boolean testTrans(Vector<RegExpr> Trans){
	  for(int i = 0; i < Trans.size(); i++){
		  if(!alphabet.contains(Trans.get(i))) return false;
		  if(!isUnobserv(Trans.get(i))) return false;
	  }
	  return true;
  }
  
  /**
   * will run a the Cluster Verifier algorithm on the automata with the parameters needed
   * @param policies  a Vector of the unobservable transitions
   * @param all_trans a Vector of all the transitions in the automata
   * @param workDir the working directory
   * @param filename this will be the output filename if the output isn't suppressed
   * @param fail_events a Vector of the failure events
   * @param suppress_output if true it will not write to a file
   * @return a string that is either "true" "false" or some error message
   */
  public String runCVerify(Vector<Vector<Transition>> policies, Vector all_trans,   String workDir,
		  				   String filename, Vector fail_events, boolean suppress_output){
	 
	  Vector<ClusterAutomata> cluster_autos = new Vector<ClusterAutomata>();
	  Vector<Transition> trans = new Vector<Transition> ();
	  String errStr = "";
	  
	  ClusterAutomata tmp_clusterAuto = new ClusterAutomata();
	  tmp_clusterAuto.reset_id();

	  for(int i = 0; i < policies.size(); i++){
		  cluster_autos.add(new ClusterAutomata(this, all_trans, policies.get(i), fail_events));
	  }
	  if(cluster_autos==null||cluster_autos.size()==0){
		  errStr = "must enter at least one agent";
		  return errStr;
	  }
	  ClusterAutomata tmp = new ClusterAutomata(cluster_autos);
	  cluster_autos.add(tmp);
	  
	  
	  Xd xd_init = new Xd(cluster_autos);
	  Vector<Xd> c_verify = new Vector<Xd>();
	  c_verify.add(xd_init);
	
	  Vector<Xd> states_out = new Vector<Xd>();
	  for(int i = 0; i < c_verify.size(); i++){
		  Xd cur_xd = ((Xd) c_verify.get(i));
		  Cluster cur_psuedo_state = ((Xd) c_verify.get(i)).getCurPsuedo(); 
		  trans = cur_psuedo_state.trans_out;
		  for(int j = 0; j < trans.size(); j++){

			  //if(!cur_xd.canTrans(((Transition) trans.get(j)).label.getEvent()) continue;
			  
			  Vector<Vector<Cluster>> tmp_vec = new Vector<Vector<Cluster>>();
			  for(int k = 0; k < cur_xd.getAutomatas().size(); k++){
				  tmp_vec.add(cur_xd.getAutomatas().get(k).R(trans.get(j), cur_xd.getCurState().get(k)));
			  }
			  boolean bad_tmp_vec = false;
			  for(int k = 0; k < tmp_vec.size(); k++){
				  if(tmp_vec.get(k)==null){ bad_tmp_vec = true; break;}
			  }
			  if(bad_tmp_vec) continue;

			  states_out = cur_xd.getXdOut(tmp_vec, ((Transition) trans.get(j)).label.getEvent());
			  for(int n = 0; n < states_out.size(); n++){
				  if(!c_verify.contains(states_out.get(n))) c_verify.add(states_out.get(n));
			  }
			  states_out.clear();
		  }

	  }

	  boolean diagnosable;
	  try{
		  diagnosable = findClusterDiagnosis(c_verify, suppress_output, workDir, filename + ".diag");
		  
		  if(!suppress_output) printfsm_xd(workDir, filename, c_verify);
	  } catch(IOException e){
		  errStr = "Unknown error has occured";
		  e.printStackTrace();
		  return errStr;} 
	  
	  tmp.reset_id();
	  if(suppress_output) return "" + diagnosable;
	  return errStr;
  }
  /**
   * This finds if the automata is diagnosible based on the xd states.
   * It also give the option of printing to a file
   * @param xd_states a vector of xd states describing the system
   * @param suppress_output true if it is not to write to a file
   * @param workDir the working directory
   * @param fileName the name of the file that should be printed to
   * @return will return true if the system is diagnosable and false if it is not
   * @throws IOException
   */	  
  private boolean findClusterDiagnosis(Vector<Xd> xd_states,
		  							   boolean suppress_output, String workDir , String fileName)
  									   throws IOException{

	  BufferedWriter out;
	  
	  //this shouldn't happen
	  if(xd_states==null || xd_states.size()==0){
		  if(!suppress_output){
			  out = new BufferedWriter(new FileWriter(workDir + fileName));
			  out.write("Empty cluster automata");
		  }
		  return false; 
		  }
	  
	  ClusterAutomata psuedo = ((Xd) xd_states.get(0)).getPsuedo();
	  

	  Vector<State> states_with_Cp_fk   = new Vector<State>();
	  Vector<State> states_with_Cp      = new Vector<State>();
	  Vector<Xd> violate_Cp_cycle    	= new Vector<Xd>();
	  Vector<Xd> violate_Cp_fk_cycle 	= new Vector<Xd>();
	  
	  Vector<Xd> f_labeled_states    	= new Vector<Xd>();
	  Vector<Xd> cycle_help          	= new Vector<Xd>();
	  boolean found_f_cycle = false;
	  for(int i = 0; i < psuedo.clusters.size(); i++){
		  Cluster c = (Cluster) psuedo.clusters.get(i);
		  if(c.Cp_cycle && !states_with_Cp.contains(c.startState))
			  states_with_Cp.add(c.startState);
		  if(c.Cp_fcycle && !states_with_Cp_fk.contains(c.startState))
			  states_with_Cp_fk.add(c.startState);
	  }
	  
	  for(int i = 0; i < xd_states.size(); i++){
		  
		  if(states_with_Cp_fk.contains(((Xd) xd_states.get(i)).getCurPsuedo().startState)) {
			  if(!violate_Cp_fk_cycle.contains((Xd) xd_states.get(i)))
				  violate_Cp_fk_cycle.add((Xd) xd_states.get(i));
		  }
		  if(states_with_Cp.contains(((Xd) xd_states.get(i)).getCurPsuedo().startState) && 
				  					!((Xd) xd_states.get(i)).getCurPsuedo().Nfk) {
			  
			  if(!violate_Cp_cycle.contains((Xd) xd_states.get(i)))
				  violate_Cp_cycle.add((Xd) xd_states.get(i));
		  }
		  if(!((Xd) xd_states.get(i)).getCurPsuedo().Nfk)
			  f_labeled_states.add((Xd) xd_states.get(i));
		  		  
	  }
	  
	  Vector<Transition> trans = new Vector<Transition>();
	  //Xd cycleHead = null;
	  for(int i  = 0; i < f_labeled_states.size(); i++){
		  Xd cur_xd = (Xd) f_labeled_states.get(i);
		  //State s = new State(cur_xd.toString());
		  trans.clear();
		  cycle_help.clear();
		  trans = (Vector) cur_xd.getTrans_out().clone();
		  if(trans==null || trans.size()==0) continue;
		  for(int j = 0; j < trans.size(); j++){
			  Xd tmp_xd = null;
			  for(int n = 0; n < f_labeled_states.size(); n++){
				  if(f_labeled_states.get(n).toString().equals(((Transition) trans.get(j)).to.name)){
					  tmp_xd = (Xd) f_labeled_states.get(n); 
					  break;
				  }
			  }
			  if(tmp_xd==null) continue;
			  if(!cycle_help.contains(tmp_xd)) cycle_help.add(tmp_xd);
			  if(tmp_xd.equals(cur_xd))found_f_cycle = true; 

		  }
		  for(int j = 0; j < cycle_help.size(); j++){
			  trans.clear();
			  trans = (Vector) ((Xd) cycle_help.get(j)).getTrans_out().clone();
			  if(trans==null || trans.size()==0) continue;
			  for(int k = 0; k < trans.size(); k++){
				  Xd tmp_xd = null;
				  for(int n = 0; n < f_labeled_states.size(); n++){
					  if(f_labeled_states.get(n).toString().equals(((Transition) trans.get(k)).to.name)){
						  tmp_xd = (Xd) f_labeled_states.get(n); 
						  break;
					  }
				  }
				  if(tmp_xd==null) continue;
				  if(!cycle_help.contains(tmp_xd)) cycle_help.add(tmp_xd);
				  if(tmp_xd.equals(cur_xd))found_f_cycle = true;
			  }	
			  if(found_f_cycle) break;
		  }
		  if(found_f_cycle) break;		  
	  }
	  
	  
	  //only applicable when writing the results
	  if(!suppress_output){  
		  out = new BufferedWriter(new FileWriter(workDir + fileName));
		  out.write("Cp_fk_cycles\n");
		  if(violate_Cp_fk_cycle.size() == 0){
			  out.write("NO FK_CYCLES FOUND");
		  }
		  if(violate_Cp_fk_cycle.size() > 5){
			  out.write("multiple cycles found: ");
			  if(states_with_Cp_fk.size() <= 10){
				  out.write("problem is when Cp is in state(s): \n");
				  for(int i = 0; i < states_with_Cp_fk.size(); i++){
					  out.write(((State) states_with_Cp_fk.get(i)).name + "  ");
				  }
			  }
			  if(states_with_Cp_fk.size() > 10){
				  out.write("\nExample of bad state: ");
				  out.write(((Xd) violate_Cp_fk_cycle.get(0)).toString());
			  }			  
		  } else if (violate_Cp_fk_cycle.size() <= 5 && violate_Cp_fk_cycle.size() > 0) {
			  out.write("Problem in state(s):  "); 
			  for(int i = 0; i < violate_Cp_fk_cycle.size(); i++){
				  out.write(((Xd) violate_Cp_fk_cycle.get(i)).toString() + "  ");
			  }
		  }
		  out.write("\n\n");

		  out.write("Cp_cycles\n");
		  if(violate_Cp_cycle.size() == 0){
			  out.write("NO STATES WITH CYCLES AND AN F LABEL FOUND");
		  }
		  if(violate_Cp_cycle.size() > 5){
			  out.write("multiple cycles found: ");
			  if(states_with_Cp.size() <= 10){
				  out.write("problem is when Cp is in state(s) and has an f label: \n ");
				  for(int i = 0; i < states_with_Cp.size(); i++){
					  out.write(((State) states_with_Cp.get(i)).name + "  ");
				  }
			  }
			  if(states_with_Cp.size() > 10){
				  out.write("\nExample of bad state: ");
				  out.write(((Xd) violate_Cp_cycle.get(0)).toString());
			  }			  
		  } else if (violate_Cp_cycle.size() <= 5 && violate_Cp_cycle.size() > 0) {
			  out.write("Problem in state(s):  "); 
			  for(int i = 0; i < violate_Cp_cycle.size(); i++){
				  out.write(((Xd) violate_Cp_cycle.get(i)).toString() + "  ");
			  }
		  }
		  out.write("\n\n");	  
	  
		  out.write("F labeled Cycles\n");
		  if(!found_f_cycle) out.write("NO F LABELED CYCLES");
		  else{
			  out.write("cycles found... to determine where the cycles are " +
					    "find the strongly\nconnected component of the output");
		  }
		  out.write("\n\n");	  
					  
		  out.close();
	  }
	  
	  if(violate_Cp_fk_cycle.size() == 0 && violate_Cp_cycle.size() == 0 && !found_f_cycle) return true;
	  
	  return false;
	  
  }
  
  private void printfsm_xd(String workDir , String fileName,  Vector<Xd> xd_states) throws IOException{
	  //String umdesExt = new UMDES_File().getExtension();
	  String umdesExt = ".fsm";
	  BufferedWriter out;
	  out = new BufferedWriter(new FileWriter(workDir + fileName + umdesExt));		

	  out.write(xd_states.size()+"\n\n");
    
	  for(int i = 0; i < xd_states.size(); i++){
    	
		  out.write(((Xd) xd_states.get(i)).toString() + "\t0\t" + 
				    ((Xd) xd_states.get(i)).getTrans_out().size() + "\n");
		  
		  Vector<Transition> trans = ((Xd) xd_states.get(i)).getTrans_out();
		  for(int j = 0; j < trans.size(); j++){
			  out.write(trans.get(j).label.expression + "\t" + 
					    trans.get(j).to.name + "\tc\to\n");		
		  }
		  out.write("\n");  	
	  }
	  out.close();
 }
  
  /**
   * finds all of the observable transitions in this automata
   * @return a Vector of Transition's
   */
  public Vector<Transition> getAllTransitions(){
	  Vector<Transition> allTrans = new Vector<Transition>();
	  Vector<Transition> stateTrans = new Vector<Transition>();
	  
	  for(int i = 0; i < states.size(); i++){
		  try{
			  stateTrans = events.getTransitions((State) states.get(i)); 
		  } catch(Exception e) { continue;}
		  for(int j = 0; j < stateTrans.size(); j++){
			  
			  if(!allTrans.contains( stateTrans.get(j)) && 
			     !isUnobserv(stateTrans.get(j).label.getEvent()))
				  	allTrans.add(stateTrans.get(j));			  
		  }	  
	  }  	  
	  return allTrans;
  }
  
  



//these methods are the ones needed for the online-min-sen-act algorithm, and an offline run of the same algorithm    
    /**
     * Will prompt the user to enter the pairs that belong to T_spec 
     * @param workDir - this is the working directory currently used
     * @throws Exception - on any error it will throw a general exception that will cause an error message
     */
      public void enterSpec(String workDir) throws Exception, IOException, InValidSpecException{
    	  
    		BufferedReader input = new BufferedReader(new FileReader(workDir + "input.spec"));
    		
    	    String str = "";
    	    State s1 = new State("");
    	    State s2 = new State("");
    	    
    	    boolean error = false;
    	    String errorStr = "";


    		while((str = input.readLine()) != null){
    			
    			str = str.trim();
    			
    			//find where white space is(space,tab)
    			int delimPosition = str.indexOf(' ');
    			if(delimPosition==-1) delimPosition = str.indexOf('\t');
    			if(delimPosition==-1) {
    				error =true;
    				errorStr = str;
    				continue;
    			}  //there are not two states
    			
    			String str1 = str.substring(0, delimPosition).trim();
    			String str2 = str.substring(delimPosition).trim();
    			
    			if(hasState(str1)&&hasState(str2)){
    				s1 = getState(str1);
    				s2 = getState(str2);
    				t_spec.addToSpec(s1, s2);
    			}
    	    	else {
    	    		error = true;
    	    		errorStr = str;
    	    		continue;
    	    	}	    	   	    	
    	    }
    	    t_spec.extend_spec(this);
    	    input.close();
    	    if(error){
    	    	t_spec.clearSpec();
    	    	throw new IOException(errorStr);
    	    }
    	       	    
    		Vector<State> initial = new Vector<State>();
    		initial.add(getInitial());
    		initial = unObservableReach(initial,null);
    		if(violate_Te_spec(initial)) 
    			throw new InValidSpecException("The initial state is in Te_spec" +
    									"\nThis FSA will always violate the spec");
    	    
      }
      /**
       * Will prompt the user to enter the pairs that belong to T_spec 
       * @param workDir - this is the working directory currently used
       * @throws Exception - on any error it will throw a general exception that will cause an error message
       */
        public void enterSpec(Vector<StatePair> spec) throws Exception, IOException, InValidSpecException{

      	    for(StatePair s: spec)
      	    	t_spec.addToSpec(s.first, s.second);
      	    
      	    t_spec.extend_spec(this);
      		Vector<State> initial = new Vector<State>();
      		initial.add(getInitial());
      		initial = unObservableReach(initial,null);
      		if(violate_Te_spec(initial)) throw new InValidSpecException("Spec is violated by the initial state");
      	    
        }
   
    /**
     * 
     * This function will call the min_sen_act algorithm and output the results to online_min_sen_act.out
     * @param workDir : the place where the file is 
     * @param r : the observed event
     * @throws IOException
     */  
    
   /* public void computeMinSenAct(String workDir, JFrame parent, RegExpr r)throws IOException{

  	    //initialize the stateSet to start at the initial state
      	if(r==null){
      		outOnlineMinSen = new BufferedWriter(new FileWriter(workDir + "online_min_sen_act.out"));
  		    orderTransitions(workDir, parent);
  		    printTSpec(workDir, outOnlineMinSen); //prints the T_spec to the file
      		Vector initial_state = new Vector();
      		initial_state.add(getInitial());
      		getStateSet().states = unObservableReach(initial_state,null);
  		    outOnlineMinSen.write("\nINITIAL\n");   		
      	}
  	    	  
      	//do the transition of event r
  	    if(r!=null){
  	    	getStateSet().clearEvents();  
  	    	setStateSet(doTransition(getStateSet(),r));
  		    outOnlineMinSen.write("\nEVENT OBSERVED: " + r.expression + "\n");
  	    }
  	    
	  
      	outOnlineMinSen.write("States-------->  " + getStateSet().outputStates() + "\n");  	//output the states before
  	    min_sen_act();
  	  	outOnlineMinSen.write("Turned off:  " + getStateSet().policyEUA() + "\n");
  	  	outOnlineMinSen.write("Kept on:     " + getStateSet().policyEA() + "\n");
  	  	outOnlineMinSen.write("Sensors off--->  " + getStateSet().outputStates() + "\n");
  	    
  	    getStateSet().states = unObservableReach(getStateSet().states,null);

    }*/
    
    /**
     * A breadth first search of all possible sets of states that the machine
     * can be in given the online_min_sen_act algorithm
     * @param workDir : path to the output file
     * @param outFSM : if the input is true, the method will remember the states so a fsm can
     * 			     be created
     * @throws IOException
     */
   /* public void runOffline(String workDir, boolean outFSM)throws IOException{
  	  
  	  Vector<Qc> working = new Vector<Qc>();  //will work like a queue
  	  boolean initialState = true;
  	  Qc tmp_Qc = new Qc();  	  
  	  RegExpr r = new RegExpr();
  	  
  	  //the regexpr will be removed if there is a counter example while running the program
  	  HashSet<RegExpr> sensorAlwaysOn  = new HashSet<RegExpr>();
  	  HashSet<RegExpr> sensorAlwaysOff = new HashSet<RegExpr>();
  	  for(int i = 0; i < alphabet.size(); i++){
  		  if(!isUnobserv(((TransitionLabel) alphabet.get(i)).getEvent())){
  			  sensorAlwaysOn.add(alphabet.get(i));
  			  sensorAlwaysOff.add(alphabet.get(i));
  		  }
  	  }
  	  
  	  //orderTransitions(workDir, parent);
  	  
  	  getStateSet().states_unchanged.clear();
  	  getStateSet().clearEvents();
  	  
  	  Vector<State> initial_state = new Vector<State>();
  	  initial_state.add(getInitial());
  	  getStateSet().states = unObservableReach(initial_state,null);	  
  	   	  
  	  working.add(getStateSet());
  	  
  	  while(working.size()!=0){
  		  
  		  setStateSet(((Qc) working.get(0)));
  		  working.remove(0);

  		  min_sen_act();  //applies the online min_sen_act algorithm on stateSet
  		  
  		  if(initialState) {getStateSet().isInitial = true; initialState = false;} //to identify the initial state later
  		  //if it already is in the Set then skip to next in queue
  		  if(possibleStateSets.contains((Qc) getStateSet())&&!getStateSet().states.isEmpty()) continue;
  		  
  		  if(getStateSet().E_a==null)continue;
  		  Iterator<RegExpr> it = getStateSet().E_a.iterator();
  		  
  		  //add a stateSet to the queue for each observable event		  
  		  while (it.hasNext()) {
  			  // Get element
  			  r = (RegExpr) it.next();
  			  tmp_Qc = doTransition(getStateSet(),r);
  			  working.add(tmp_Qc);
  			  sensorAlwaysOff.remove(r);
  			  if(outFSM) getStateSet().destStates.put(r, tmp_Qc);  //for constructing fsm if needed
  		  }
  		  possibleStateSets.add(getStateSet());
  		  
  		  if(getStateSet().E_ua==null||getStateSet().E_ua.size()==0)continue;
  		  it = getStateSet().E_ua.iterator();	  
  		  while (it.hasNext()) {
  			  r = (RegExpr) it.next();
  			  sensorAlwaysOn.remove(r);
  		  }
  		  

  	  }
  	  printOffline(workDir, sensorAlwaysOn, sensorAlwaysOff); //write the output to the online_min_sen_act
  	  if(outFSM) printFSM(workDir, true);
  	  clearSpec();
  	  sensorAlwaysOn.clear();
  	  sensorAlwaysOff.clear();
    }*/
    
   /* public void createMinSenFSM(String workDir, Vector trans_unseen) throws IOException{
    	
    	Vector<String> activeSensors = new Vector<String>();
    	clearSpec(); //makes sure everything is clean
    	
		BufferedReader input = new BufferedReader(new FileReader(workDir + "min_sen_act.out"));	
	    String str = "";

		while((str = input.readLine()) != null){			
			str = str.trim();
			if(str.equals("Confusable state pairs:")){
				//get rid of junk formatting data in file
				activeSensors.remove(activeSensors.size()-1);
				activeSensors.remove(0);
				activeSensors.remove(0);
				break;
			}
			activeSensors.add(str);

		}
		input.close();
		
    	
    	HashMap<State,HashSet<RegExpr>> stateMap  = new HashMap<State,HashSet<RegExpr>>();
    	HashSet<RegExpr> regexpInState; // = new HashSet();
    	State s = new State("");
    	RegExpr r = new RegExpr();
    	
    	
    	String[] in;
    	for(int i = 0; i < activeSensors.size(); i++){
        	boolean stateAdded = true;
    		in = ((String) activeSensors.get(i)).split(" ");
    		for(int j = 0; j < in.length; j++){
    			if(in[j].equals("")) continue;
    			if(!stateAdded){
    				r = new RegExpr(in[j]);
    				if(!stateMap.containsKey(s)){
    					regexpInState = new HashSet<RegExpr>();
    					regexpInState.add(r);
    					stateMap.put(s, regexpInState);
    				}
    				else{
    					regexpInState = stateMap.get(s);
    					regexpInState.add(r);
    				}
    			}
    			if(stateAdded){
        			s = new State(in[j]);
        			stateAdded = false;
    			}
    		}
    	}
    	
    	Vector<Qc> working = new Vector<Qc>();  //queue find all the possible states to construct fsm   	
    	Vector<RegExpr> seenEvents = new Vector<RegExpr>();
    	Qc tmp_Qc = new Qc();
    	Qc new_Qc = new Qc();
    	
    	State initial = getInitial();
    	tmp_Qc.states.add(initial);
    	tmp_Qc.states = unObservableReach(tmp_Qc.states, null);
    	
    	//tmp_Qc.states =;
    		
    	boolean initialState = true;
    	working.add(tmp_Qc);
    	  
    	while(working.size()!=0){
    		 
    		if(seenEvents!=null) seenEvents.clear();
    		
    		tmp_Qc = ((Qc) working.get(0));
        	tmp_Qc.states = unObservableReach(tmp_Qc.states, null);        	
    		tmp_Qc.states_unchanged = new Vector<State>(tmp_Qc.states); //so equality test of Qc in sets works 
    		
    		working.remove(0);

    		//put Unoservable reach in the min_sen expansion function
    		min_sen_expansion(tmp_Qc.states,stateMap,seenEvents, tmp_Qc.E_ua, trans_unseen);
    		for(int i = 0; i < seenEvents.size(); i++){ tmp_Qc.E_a.add(seenEvents.get(i));}


    		if(initialState) {tmp_Qc.isInitial = true; initialState = false;} //to identify the initial state later
    		//if it already is in the Set then skip to next in queue
    		if(possibleStateSets.contains((Qc) tmp_Qc)&&!tmp_Qc.states.isEmpty()) continue;
    		  
    		  
    		//add a stateSet to the queue for each observable event		  
    		for(int i = 0; i < seenEvents.size(); i++){
    			r =  seenEvents.get(i);
    			new_Qc = doTransition(tmp_Qc,r);  
    			working.add(new_Qc);
    			tmp_Qc.destStates.put(r, new_Qc);  //for constructing fsm if needed
    		}
    		possibleStateSets.add(tmp_Qc);
    		  
    	}
    	
    	printFSM(workDir, false);

    	clearSpec();
    	
    }*/
    /*private void min_sen_expansion(Vector curStates, HashMap map, Vector seen, Set E_ua, Vector unseen_total){
    	
    	//curStates = unObservableReach(curStates, null);
    	
    	RegExpr r = new RegExpr();
    	HashSet set = new HashSet();
    	//Vector seen = new Vector();
    	Vector trans = new Vector();
    	State s = new State("");
    	for(int i = 0; i < curStates.size(); i++){
    		try{
    			trans = events.getTransitions(((State) curStates.get(i)));
    		}catch(Exception e){}
    		if(trans==null) continue;
    		for(int j = 0; j < trans.size(); j++){
    			r = ((Transition) trans.get(j)).getLabel().getEvent();
    			if(isUnobserv(r)) {
    				if(!unseen_total.contains(((Transition) trans.get(j))))
    					unseen_total.add(((Transition) trans.get(j)));
    				continue;
    			}
    			if(map.containsKey(((State) curStates.get(i)))){
    				set = (HashSet) map.get(((State) curStates.get(i)));
    				if(set.contains(r)){
    					if(!seen.contains(r)) seen.add(r);
    					continue;
    				}
    				else{
    					//if(!unseenEvents.contains(r)) unseenEvents.add(r);
        				if(!unseen_total.contains(((Transition) trans.get(j))))
        					unseen_total.add(((Transition) trans.get(j)));
    					s = ((Transition) trans.get(j)).to;
    					E_ua.add(new RegExpr(((State) curStates.get(i)).name + "-" + r.expression));
    					if(!curStates.contains(s)){
    						curStates.add(s);
    				    	curStates = unObservableReach(curStates, null);   						
    					}
    				}
    			}
    			else{
    				//can skip checking the map if it contains no key for the state
    				for(;j < trans.size();j++){
        				if(!unseen_total.contains(((Transition) trans.get(j))))
        					unseen_total.add(((Transition) trans.get(j)));
    	    			r = ((Transition) trans.get(j)).getLabel().getEvent();
    					if(isUnobserv(r)){j++; continue;}
    					//if(!unseenEvents.contains(r)) unseenEvents.add(r);
    					s = ((Transition) trans.get(j)).to;
    					E_ua.add(new RegExpr(((State) curStates.get(i)).name + "-" + r.expression));
    					if(!curStates.contains(s)){
    						curStates.add(s);
    				    	curStates = unObservableReach(curStates, null);    						
    					}
    				}  				
    			}
    		}
    	}
    	//it has now expanded to contain all of the states from the unactivated sensors
    	

    	   	
    	//return curStates;
    }*/

    /**
     * Empties t_spec and all that depend on it
     */
    public void clearSpec(){
  	  getStateSet().clearAll();
  	  possibleStateSets.clear();
  	  t_spec.clearSpec();
  	  getTrans_order().clear();
  	  try{
  		  outOnlineMinSen.close();
  	  }catch(Exception e){}
  	  outOnlineMinSen = null;
    }
    
    /**
     * 
     * @param curStates : the current set of states
     * @param r : the expression being examined
     * @return true iff at least one state has this transition to any other state
     */
    public boolean canTransition(Qc curStates, RegExpr r){
  	 
  	  //look at each state and if any of them uses the transition then it is true
  	  for(int i = 0; i< curStates.states.size(); i++){
  		  if(events.hasTransition((State) curStates.states.get(i), r)) return true; 
  	  } 	  
  	  return false;
    }
    
    /**
     * drag and drop user interface to enter the transitions in the order that you want to 
     * consider them when using the online sensor activation algorithms.
     * @param workDir the working directory of the .transorder file
     * @throws if the input file is incomplete it will throw a general Exception
     */
     /*public void orderTransitions(String workDir, JFrame parent)throws IOException{
    	  
        getTrans_order().clear();
   	     	    
        for(int i = 0; i<alphabet.size(); i++){
        	if(isUnobserv(((TransitionLabel) alphabet.get(i)).getEvent())){
        		getTrans_order().add(((TransitionLabel) alphabet.get(i)).getEvent());
        		continue;	    	
        	}
        }
        Vector events = getObservableEvents();
        EventList ev = new EventList(events, parent, false);
   	    if(ev.getEventOrder()==null||ev.getEventOrder().size()==0) throw new IOException();

   	    EventOrder.clear();
   	    for(int i = 0; i<ev.getEventOrder().size(); i++){
    	   	getTrans_order().add(new RegExpr((String) ev.getEventOrder().get(i)));
    	   	EventOrder.add(((String) ev.getEventOrder().get(i)));
    	   }

      }*/
     
    /* public void orderAllTransition(String workDir, JFrame parent)throws IOException{
    	 //make sure old files are gone
  	  	  File f = new File(workDir + "H.transorder");
   	  	  f.delete();    	 
    	
    	 //initialize new order else use already entered
    	 if(totalOrder==null||totalOrder.size()==0){
    		 totalOrder = new Vector();
    		 Vector seen_trans = new Vector();
    		 String str_state_in   = "";
    		 String str_transition = "";
    		 Vector v_trans  = new Vector();
    		 Vector v_states = new Vector();
    		 v_states = getStates();
    		 for(int i = 0; i < v_states.size(); i++){
    			 try{
        			 v_trans = events.getTransitions((State) v_states.get(i));    				 
    			 }
    			 catch(Exception e){ continue;}
    			 str_state_in = ((State) v_states.get(i)).name;
    			 seen_trans.clear();
    			 for(int j = 0; j < v_trans.size(); j++){
    				 if(isUnobserv(((Transition) v_trans.get(j)).label.getEvent())) continue;
    				 str_transition = ((Transition) v_trans.get(j)).label.expression;
    				 
    				 if((seen_trans!=null||seen_trans.size()!=0)&&
    						 seen_trans.contains(((Transition) v_trans.get(j)).label.expression)){
    					 continue;
    				 }
    				 
    				 
    				 
    				 seen_trans.add(((Transition) v_trans.get(j)).label.expression);
    				 totalOrder.add(0,str_state_in + "      " +  str_transition);// + "\t" + str_state_out);
    			 }
    		 }
    	 }
    	 EventList e = new EventList(totalOrder, parent, true);
    	 totalOrder = e.getEventOrder();
    	 
    	 if(totalOrder==null||totalOrder.size()==0){ throw new IOException();}
    	 
    	 BufferedWriter transorder = new BufferedWriter(new FileWriter(workDir + "H.transorder"));
    	 for(int i = 0; i < totalOrder.size(); i++){
    		 transorder.write(((String) totalOrder.get(i)) + "\n");
    	 }
    	 transorder.close();
    	 
    	 
     }*/
    
    /**
     * updates the current set of states to be a new set of states that has gone
     * through transition r
     * @param curStates : current set of states
     * @param r : event observed
     * @return the new set of states
     */
    
    public Qc doTransition(Qc curStates, RegExpr r){
  	   
  	  Vector tmp_trans = new Vector();
  	  Qc new_Qc = new Qc();
  	  
  	  for( int i = 0; i< curStates.states.size(); i++){
  		  if(events.hasTransition(((State) curStates.states.get(i)), r)){
  			  tmp_trans =  events.getTransition(((State) curStates.states.get(i)), r);
  		  }
  		  if(tmp_trans==null)continue;
  		  for(int j = 0; j<tmp_trans.size(); j++){
  			  if(!new_Qc.states.contains(((Transition) tmp_trans.get(j)).to)) 
  				  new_Qc.states.add(((Transition) tmp_trans.get(j)).to);
  		  }
  	  }

  	  new_Qc.states = unObservableReach(new_Qc.states,null);
  	  
  	  return new_Qc;
    }
      
    
  /**
   * 
   * Looks at the vector containing the possible states a machine is in
   * and updates it to contain those states plus that set's unobservable 
   * reach, allows for the input of one observable event to also be
   * considered unobservable (turning off the sensor).  If this is 
   * not needed set r to null when calling
   * 
   * @param set_states the states that the machine is in
   * @param r sensor to be turned off
   * @return the set of states along with its unobservable reach
   */
    public Vector<State> unObservableReach(Vector<State> set_states, RegExpr r){

  	    Vector transition_vec = new Vector();
  	    State state = new State("");

  	    if(set_states==null)return null;
  	    
  	    //iterate through the states
  	    for(int i = 0; i<set_states.size(); i++){

  	        state = (State) set_states.get(i);
  	        
  	        try{transition_vec = events.getTransitions(state);}
  	        catch(Exception e){ }	        
  	        if(transition_vec==null)continue;


          	for(int j = 0; j< transition_vec.size(); j++){
          		
          		//conditions needed to look at to add to set_states
          		boolean isUnobserv, isR, isOff, notInSet;
          		isUnobserv = isUnobserv(((Transition)transition_vec.get(j)).getLabel().getEvent());
          		isR   = ((Transition)transition_vec.get(j)).getLabel().getEvent().equals(r);
          		isOff = (r!=null&&getStateSet().E_ua.contains(((Transition)transition_vec.get(j)).getLabel().getEvent()));
          		notInSet   = !set_states.contains(((Transition)transition_vec.get(j)).to);
          		
          		if((isUnobserv || isR || isOff) && notInSet){
          /*if((isUnobserv(((Transition)transition_vec.get(j)).getLabel().getEvent())||  //look at unobservable
          	((Transition)transition_vec.get(j)).getLabel().getEvent().equals(r)||  //look at r
          	(r!=null&&stateSet.E_ua.contains(((Transition)transition_vec.get(j)).getLabel().getEvent())))&&  
          	!set_states.contains(((Transition)transition_vec.get(j)).to)){*/
  	        			
          			set_states.add(((Transition)transition_vec.get(j)).to);	        		
          		}
          	}
          	
  	    }

  	    return set_states;
    }
    
    
    /**
     * tests to see if the states in the vector are illegal according to T_spec
     * @param v The vector of states that is to be tested
     * @return true if it is illegal, false if it is a good set of states
     */
    private boolean violate_Te_spec(Vector v){
  	  
  	  if(v==null)return false;
  	  Vector illegal= t_spec.violate_vector();
  	  
  	  for(int i = 0; i<v.size();i++){
  		  for(int j = 0; j<v.size(); j++){
  			  StatePair s = new StatePair(((State) v.get(i)),((State) v.get(j)));
  			  if(illegal.contains(s)) return true;
  		  }
  	  }	  
  	  return false;
    }   
    
    /**
     * changes the stateSet of this Fa to include the activation policy in its
     * hashSets (E_a, E_ua, E_uo). It then changes the stateSet.states to include
     * the added states by the turn off sensors from the activation policy.
     * StateSet.states_unchanged remains the same as states before this function.
     * 
     */
    /*private void min_sen_act(){
  	  
  	  Vector<State> expanded_Qc = new Vector<State>();
  	  
  	  //deep copy
  	  getStateSet().states_unchanged.clear();
  	  getStateSet().clearEvents();
  	  for(int k = 0; k<getStateSet().states.size(); k++){
  		  expanded_Qc.add(getStateSet().states.get(k));
  		  if(!getStateSet().states_unchanged.contains(getStateSet().states.get(k)))
  			  getStateSet().states_unchanged.add(getStateSet().states.get(k));
  	  }

  	  //for all transitions
  	  for(int i = 0; i<getTrans_order().size(); i++){

  		  if(!isUnobserv(((RegExpr) getTrans_order().get(i)))){

  			  expanded_Qc = unObservableReach(expanded_Qc,((RegExpr) getTrans_order().get(i)));
  			  if(!violate_Te_spec(expanded_Qc)){
  				  getStateSet().E_ua.add(((RegExpr) getTrans_order().get(i))); 
  				  //add all states after Qc ends (set Qc = expanded_Qc)
  				  while(getStateSet().states.size()!=expanded_Qc.size()){
  					  getStateSet().states.add(expanded_Qc.get(getStateSet().states.size()));
  				  }
  			  }
  			  else{
  				  getStateSet().E_a.add(((RegExpr) getTrans_order().get(i)));  
  				  //remove all states after Qc ends (set expanded_Qc = Qc)
  				  while(getStateSet().states.size()!=expanded_Qc.size()){
  					  expanded_Qc.remove(getStateSet().states.size());
  				  }
  			  }  			  
  		  }
  		  else getStateSet().E_uo.add(((RegExpr) getTrans_order().get(i)));  //it is unobservable

  	  }
  	  return;
    }*/
    
    /**
     * Does all of the needed printing to file in runOffline
     * @param workDir : the path to the output
     * @throws IOException
     */
   /* private void printOffline(String workDir, HashSet On, HashSet Off) throws IOException{

        BufferedWriter out = new BufferedWriter(new FileWriter(workDir + "offline_min_sen_act.out"));
  	    printTSpec(workDir, out);  

  	  	Iterator it = possibleStateSets.iterator();
  	  	while (it.hasNext()) {
  	  		// Get element
  	  		Qc tmp_Qc = (Qc) it.next();
  	  		if(tmp_Qc.states.isEmpty()||tmp_Qc.states_unchanged.isEmpty()) continue;

  	  		out.write("\nStates-------->  " + tmp_Qc.outputStatesUnchanged() + "\n");  //output the states before
  	  		out.write("Turned off:  " + tmp_Qc.policyEUA() + "\n");		//output the events turned off
  	  		out.write("Kept on:     " + tmp_Qc.policyEA() + "\n");		//output the events that can be observed
  	  		out.write("Sensors off--->  " + tmp_Qc.outputStates() + "\n");	//output the states after
  	      	        		
  	  	}    
  	  	if(!On.isEmpty()){
  	  		out.write("\n");
  	  		it = On.iterator();
  	  		out.write("SENSORS ALWAYS ON:  {");
  	  		while(it.hasNext()){
  	  			out.write(((RegExpr) it.next()).getExpression());
  	  			if(it.hasNext()) out.write(", ");
  	  			else out.write("} ");
  	  		}
  	  		out.write("\n");
  	  	}
  	  	if(!Off.isEmpty()){
  	  		out.write("\n");
  	  		it = Off.iterator();
  	  		out.write("SENSORS ALWAYS OFF:  {");
  	  		while(it.hasNext()){
  	  			out.write(((RegExpr) it.next()).getExpression());
  	  			if(it.hasNext()) out.write(", ");
  	  			else out.write("} ");
  	  		}
  	  		out.write("\n");
  	  	}
        out.close();
    }*/
    
   /* private void printFSM(String workDir, boolean offline) throws IOException{
 
    	//String umdesExt = new UMDES_File().getExtension();
    	String umdesExt = ".fsm";
    	BufferedWriter out;
        if(offline) out = new BufferedWriter(new FileWriter(workDir + "offline_sensor_activation" + umdesExt));
        else out = new BufferedWriter(new FileWriter(workDir + "min_sen_act" + umdesExt));
        
        Qc workingQc = new Qc();
        Qc destQc = new Qc();
    	RegExpr r = new RegExpr();
        String output = "";
        int position;
        
        // it is possible that the stateSets will be in a different order, 
        //so they must be checked if they are equal to orderalready entered stateSets
        Vector seenStateUnchanged = new Vector();
        Vector seenStateChanged = new Vector();
        
        out.write(possibleStateSets.size() +"\n\n");

        //add the stateSets to the seen stateSet Vector, so transition can use this form
        Iterator it_stateSet = possibleStateSets.iterator();
        while(it_stateSet.hasNext()){
        	workingQc = (Qc) it_stateSet.next();
        	seenStateUnchanged.add(workingQc.states_unchanged);
        	seenStateChanged.add(workingQc.states);
        }
        
        it_stateSet = possibleStateSets.iterator();
        while(it_stateSet.hasNext()){
        	workingQc = (Qc) it_stateSet.next();
        	if(workingQc.isInitial) break;
        }
        String str;
        str = "(" + workingQc.outputStatesUnchanged()+ "," + workingQc.outputStates()+ 
              "," + workingQc.policyEUA() + ")\t0\t" + workingQc.E_a.size() + "\n";

        
        //must be in special format
        str = str.replaceAll(" ", "");
        str = str.replaceAll("[{]", "[");
        str = str.replaceAll("[}]", "]"); 
        out.write(str);
        
        //seenStateSets.add(workingQc.states_unchanged);
        
        Iterator it_trans = workingQc.E_a.iterator();
        while(it_trans.hasNext()){
        	r = (RegExpr) it_trans.next();
        	output = "";
        	output = output + r.expression + "\t";
        	destQc = (Qc) workingQc.destStates.get(r);
        	
        	position = lookAtStateSets(seenStateUnchanged, destQc.states_unchanged);
	        if(position!=-1){
	        	destQc.states_unchanged = (Vector) seenStateUnchanged.get(position);
	        	destQc.states = (Vector) seenStateChanged.get(position);
	        }	        

            str =  "("+destQc.outputStatesUnchanged()+ "," + 
                       destQc.outputStates()+ "," + destQc.policyEUA() + ")";

	        //done independently of output just incase the transition for some reason has '{' or '}'
	        str = str.replaceAll(" ", ""); 
	        str = str.replaceAll("[{]", "[");
	        str = str.replaceAll("[}]", "]");
	        output = output + str + "\tc\to\n";

		    out.write(output);

        }
        out.write("\n");
        //finished initial now output the rest
        
        it_stateSet = possibleStateSets.iterator();
        while(it_stateSet.hasNext()){
        	workingQc = (Qc) it_stateSet.next();
        	if(workingQc.isInitial) continue;

            str = "("+workingQc.outputStatesUnchanged()+ "," + workingQc.outputStates()+ "," + 
                      workingQc.policyEUA() + ")\t0\t" + workingQc.E_a.size() + "\n";
            
            //must be in special format
            str = str.replaceAll(" ", "");
            str = str.replaceAll("[{]", "[");
            str = str.replaceAll("[}]", "]"); 
            out.write(str);
            
            it_trans = workingQc.E_a.iterator();
            while(it_trans.hasNext()){
            	r = (RegExpr) it_trans.next();
            	output = "";
            	output = output + r.expression + "\t";
            	destQc = (Qc) workingQc.destStates.get(r);
            	
            	position = lookAtStateSets(seenStateUnchanged, destQc.states_unchanged);
    	        if(position!=-1){
    	        	destQc.states_unchanged = (Vector) seenStateUnchanged.get(position);
    	        	destQc.states = (Vector) seenStateChanged.get(position);
    	        }

                str =  "("+destQc.outputStatesUnchanged()+ "," + 
                           destQc.outputStates()+ "," + destQc.policyEUA() + ")";
  
    	        //done independently of output just incase the transition for some reason has '{' or '}'
    	        str = str.replaceAll(" ", ""); 
    	        str = str.replaceAll("[{]", "[");
    	        str = str.replaceAll("[}]", "]");
    	        output = output + str + "\tc\to\n";
    	        out.write(output);
            }    
            out.write("\n");
        }
        
        out.write("\n\n");
    	
    	out.close();
    }*/
  
    /*private int lookAtStateSets(Vector seen, Vector lookAt){
    	
    	if(seen==null||seen.size()==0) return -1;
    	Vector tmp = new Vector();
    	
    	boolean identical;
    	boolean equal;
 
    	//identical
    	for(int i = 0; i< seen.size(); i++){
    		tmp = (Vector) seen.get(i);
    		if(tmp.size()!=lookAt.size()) continue;
			//if identical there is no changes
    		identical = true;
			for(int j = 0; j< lookAt.size(); j++){
				if(!((State) tmp.get(j)).equals((State) lookAt.get(j))){identical = false; break;}
			}
			if(identical) return -1;
    	}
    	
    	//equal
    	for(int i = 0; i< seen.size(); i++){
    		tmp = (Vector) seen.get(i);
    		//this will be the case for most
    		if(tmp.size()!=lookAt.size()) continue;

    		equal     = true;
    		for(int j = 0; j< lookAt.size(); j++){
    			if(!tmp.contains((State) lookAt.get(j))){equal = false; break;}
    		}
    		//looked through every element and didn't break
    		if(equal) return i;
    	}
    	
    	//if haven't returned true then return false
    	return -1;
    }*/
    
    /**
     * prints the expanded te_spec 
     * @param workDir : path to the output
     * @throws IOException
     */
    /*private void printTSpec(String workDir, BufferedWriter output) throws IOException{
  	  	  
  	  Vector v = t_spec.violate_vector();
  	  output.write("t_spec expanded:\n");
  	  for(int i = 0; i<v.size();i++){
  		  if((i % 6)==0&&i!=0) output.write("\n");
  		  if(i+1!=v.size())
  			  output.write(" { " + ((StatePair) v.get(i)).first.name + ", "
  					  + ((StatePair) v.get(i)).second.name + "} ,");
  	  }
  	  //don't want a comma for the last one
  	  output.write(" { " + ((StatePair) v.get(v.size()-1)).first.name + ", " 
  			  		  + ((StatePair) v.get(v.size()-1)).second.name + "}  \n");
  		  
  	  output.write("------------------------------------------------------------");
    }*/
 
    /**
     * will print the string of RegExpr that the input contains, this is  
     * printed to the end of the file specified, written by earlier print methods
     * @param v : a vector of RegExpr
     * @param out : the bufferedwriter for the output file
     * @throws IOException
     */
    public void printInputString(final Vector<RegExpr> v, BufferedWriter out)throws IOException{
  	  
    	out.write("\nINPUT STRING: \n");
    	for(int i = 0; i< v.size(); i++){
    		if((i % 45)==0) out.write("\n");
    		out.write(v.get(i).expression + " ");		  
    	}
    	out.write("\n");
    	out.close();
    }
	public void setStateSet(Qc stateSet) {
		this.stateSet = stateSet;
	}
	public Qc getStateSet() {
		return stateSet;
	}
	public void setTrans_order(Vector<RegExpr> trans_order) {
		this.trans_order = trans_order;
	}
	public Vector<RegExpr> getTrans_order() {
		return trans_order;
	}
	public void setTSpec(T_spec t_spec) {
		this.t_spec = t_spec;
	}
	public T_spec getTSpec() {
		return t_spec;
	}
	public Set<Qc> getPossibleStateSets(){
		return possibleStateSets;
	}


};

